package com.elphel.imagej.cuas;

import java.util.Arrays;
import java.util.concurrent.atomic.AtomicInteger;

import com.elphel.imagej.common.ShowDoubleFloatArrays;
import com.elphel.imagej.readers.ImagejJp4Tiff;
import com.elphel.imagej.tileprocessor.ImageDtt;
import com.elphel.imagej.tileprocessor.OpticalFlow;
import com.elphel.imagej.tileprocessor.QuadCLT;
import com.elphel.imagej.tileprocessor.QuadCLTCPU;

import ij.ImagePlus;

public class CorrectionFPN {
		final QuadCLT quadCLT;
		public CorrectionFPN(QuadCLTCPU quadCLT) {
			this.quadCLT = (QuadCLT) quadCLT;
		}
	
		  public void setApplyFPN(
				  QuadCLT []    quadCLTs,
				  double [][][] fpn) { // can be null to reset
			  quadCLT.setImageData(fpn);
			  quadCLT.setFPN(fpn);
			  for (int nscene = 0; nscene < quadCLTs.length; nscene++) if (quadCLTs[nscene] != null){
				  quadCLTs[nscene].setFPN(fpn);
				  quadCLTs[nscene].applyFPN();
			  }
			  return;
		  }
		  
		  
		  
		  public void applyFPN(
				  double [][][] fpn) {
			  if (quadCLT.getAppliedFPN() != fpn) {
				  if (quadCLT.getImageData() == null) {
					  System.out.println("applyFPN (): image_data==null");
					  return;
				  }
				  final double [][][] image_data = quadCLT.getImageData();
				  final double [][][] image_fpn_applied = quadCLT.getAppliedFPN();
				  final Thread[] threads = ImageDtt.newThreadArray(ImageDtt.THREADS_MAX);
				  final AtomicInteger ai = new AtomicInteger(0);
				  for (int ithread = 0; ithread < threads.length; ithread++) {
					  threads[ithread] = new Thread() {
						  public void run() {
							  for (int nChn = ai.getAndIncrement(); nChn < image_data.length; nChn = ai.getAndIncrement()) {
								  for (int ncol = 0; ncol < image_data[nChn].length; ncol++) {
									  for (int npix = 0; npix < image_data[nChn][ncol].length; npix++) {
										  if (image_fpn_applied != null) {
											  image_data[nChn][ncol][npix] += image_fpn_applied[nChn][ncol][npix];
										  }
										  if (fpn != null) {
											  image_data[nChn][ncol][npix] -= fpn[nChn][ncol][npix]; 
										  }
									  }
								  }
							  }
						  }
					  };
				  }
				  ImageDtt.startAndJoin(threads);
				  quadCLT.setFpnApplied(fpn);
				  quadCLT.setHasNewImageData(true);
			  }
			  return;
		  }
		  
//		  pu
		  
		  
		  
		  

		  public ImagePlus saveShowFPN(
				  double [][][] fpn,
				  int           width,
				  boolean       save,
				  boolean       show) {
			  ImagePlus imp =  CorrectionFPN.showFPN(
					  fpn, // double [][][] fpn,
					  width, // int     width,
					  show); // boolean show)
			  if (save && (imp != null)) {
				  quadCLT.saveImagePlusInModelDirectory(
						  QuadCLT.CENTER_FPN_SUFFIX, // String      suffix, // null - use title from the imp
							imp); // ImagePlus   imp)
			  }
			  return imp;
		  }
		  
		  public ImagePlus saveShowRowCol(
				  final double [][][] rows,    // [scene][sensor][width]
				  final double [][][] cols,    // [scene][sensor][height]
				  final boolean       save,
				  final boolean       show) {
			  ImagePlus imp = saveShowRowCol(
					  rows,    // final double [][][] rows,    // [scene][sensor][width]
					  cols,    // final double [][][] cols,    // [scene][sensor][height]
			       show); //final boolean       show)
			  if (save && (imp != null)) {
				  quadCLT.saveImagePlusInModelDirectory(
						  QuadCLT.CENTER_ROWCOL_SUFFIX, // String      suffix, // null - use title from the imp
							imp); // ImagePlus   imp)
			  }
			  return imp;
		  }
		  
		  
		  public static ImagePlus saveShowRowCol(
				  final double [][][] rows,    // [scene][sensor][width]
				  final double [][][] cols,    // [scene][sensor][height]
				  final boolean       show) {
			  final int num_scenes =   rows.length;
			  final int num_sens =     rows[0].length;
			  final int width =        rows[0][0].length;
			  final int height =       cols[0][0].length;
			  final int  img_width =  Math.max(width,  height);
			  final int  img_height = num_scenes;
			  final String [] titles_top = {"rows","columns"};
			  final String [] titles = new String [num_sens];
			  double [][][] data = new double [titles_top.length][num_sens][img_width*img_height];
			  for (int nsens = 0; nsens < num_sens; nsens++) {
				  titles[nsens]=String.format("sensor-%02d",nsens);
				  for (int n = 0; n < data.length; n++) {
					  Arrays.fill(data[n][nsens], Double.NaN);
				  }
				  
			  }
			  final Thread[] threads = ImageDtt.newThreadArray();
			  final AtomicInteger ai = new AtomicInteger(0);
			  for (int ithread = 0; ithread < threads.length; ithread++) {
				  threads[ithread] = new Thread() {
					  public void run() {
						  for (int nsens = ai.getAndIncrement(); nsens < num_sens; nsens = ai.getAndIncrement()) {
							  titles[nsens]=String.format("sensor-%02d",nsens);
							  for (int n = 0; n < data.length; n++) {
								  Arrays.fill(data[n][nsens], Double.NaN);
							  }
							  for (int nscene = 0; nscene < num_scenes; nscene++) {
								  System.arraycopy(
										  rows[nscene][nsens],
										  0,
										  data[0][nsens],
										  nscene*img_width,
										  width);
								  System.arraycopy(
										  cols[nscene][nsens],
										  0,
										  data[1][nsens],
										  nscene*img_width,
										  height);
							  }
						  }
					  }
				  };
			  }		      
			  ImageDtt.startAndJoin(threads);
			  String       title = "IMAGE"+QuadCLT.CENTER_ROWCOL_SUFFIX;
			  ImagePlus imp = ShowDoubleFloatArrays.showArraysHyperstack(
					  data,        // double[][][] pixels, 
					  img_width,       // int          width, 
					  title,       // String       title, "time_derivs-rt"+diff_time_rt+"-rxy"+diff_time_rxy,
					  titles,      // String []    titles, // all slices*frames titles or just slice titles or null
					  titles_top,  // String []    frame_titles, // frame titles or null
					  show);      // boolean      show)
			  imp.setProperty("numScenes",   num_scenes+"");
			  imp.setProperty("numSensors",  num_sens+"");
			  imp.setProperty("width",       width+"");
			  imp.setProperty("height",      height+"");
			  ImagejJp4Tiff.encodeProperiesToInfo(imp);
			  return imp;
		  }
		  

		
		
	  public double [][][] readImageFPN (
			  int sens_mask){

		  int [] wh = new int [2];
		  double [][] fpn_data2 = quadCLT.readDoubleArrayFromModelDirectory(
				  QuadCLT.CENTER_FPN_SUFFIX, // String      suffix,
				  0,                 // int         num_slices, // (0 - all)
				  wh); // int []      wh)
		  if (fpn_data2 == null) {
			  return null;
		  }
		  int num_sensors = quadCLT.getNumSensors();
		  int used_sensors = 0;
		  for (int nsens = 0; nsens < num_sensors; nsens++) if ( (sens_mask & (1 << nsens)) != 0) {
			  used_sensors++;
		  }
		  int num_colors = fpn_data2.length/used_sensors;
		  double [][][] fpn = new double [num_sensors][num_colors][];
		  int isens = 0;
		  for (int nsens = 0; nsens < num_sensors; nsens++) if ( (sens_mask & (1 << nsens)) != 0) {
			  for (int ncol = 0; ncol < num_colors; ncol++) {
				  fpn[nsens][ncol] = fpn_data2[ncol * used_sensors + isens];
			  }
			  isens++;
		  }
		  return fpn;
	  }

	  public double [][][][] readImageRowCol () {
		  int [] wh = new int [2];
		  ImagePlus imp = quadCLT.readImagePlusFromModelDirectory(
				  QuadCLT.CENTER_ROWCOL_SUFFIX); // String      suffix,
		  if (imp == null) {
			  return null;
		  }
		  ImagejJp4Tiff.decodeProperiesFromInfo(imp);
		  double [][] rowcol_data2 = QuadCLT.readDoubleArray(
				  imp,       // ImagePlus   imp,
				  0,         // 	int         num_slices, // (0 - all)
				  wh);       // int []      wh); // int []      wh)
		  if (rowcol_data2 == null) {
			  return null;
		  }
		  final int img_width = wh[0];
		  final int num_sens =   rowcol_data2.length / 2;
		  final int num_scenes = Integer.parseInt((String) imp.getProperty("numScenes"));
		  final int num_sens0 =   Integer.parseInt((String) imp.getProperty("numSensors"));
		  final int width =      Integer.parseInt((String) imp.getProperty("width"));
		  final int height =     Integer.parseInt((String) imp.getProperty("height"));
		  final double [][][] rows = new double[num_scenes][num_sens][width];
		  final double [][][] cols = new double[num_scenes][num_sens][height];
		  final Thread[] threads = ImageDtt.newThreadArray();
		  final AtomicInteger ai = new AtomicInteger(0);
		  for (int ithread = 0; ithread < threads.length; ithread++) {
			  threads[ithread] = new Thread() {
				  public void run() {
					  for (int nsens = ai.getAndIncrement(); nsens < num_sens; nsens = ai.getAndIncrement()) {
						  for (int nscene = 0; nscene < num_scenes; nscene++) {
							  System.arraycopy(
									  rowcol_data2[2 * nsens + 0],
									  nscene*img_width,
									  rows[nscene][nsens], // data[0][nsens],
									  0,
									  width);
							  System.arraycopy(
									  rowcol_data2[2 * nsens + 1],
									  nscene*img_width,
									  cols[nscene][nsens], // data[1][nsens],
									  0,
									  height);
						  }
					  }
				  }
			  };
		  }		      
		  ImageDtt.startAndJoin(threads);
		  return new double [][][][] {rows,cols};
	  }

	  
	  
	  public ImagePlus debugFPN(
			  QuadCLT []     quadCLTs,
			  double [][][]  fpn,
			  int []         range,
			  int            nsens,
			  double         um_sigma,
			  double         max_abs, //  =        100.0;
			  double         weight_outlier, // = 0.001;
			  boolean        show) {
		  int fpn_width =quadCLT.getTilesX()*quadCLT.getTileSize(); // see if center_CLT can be used
		  String dbg_title = quadCLT.getImageName()+"-DEBUG_FPN_ROWCOL_SENS_"+nsens+"-SIGMA"+um_sigma+"-MAXABS"+max_abs;
		  ImagePlus imp = debugFPN(
				  quadCLTs,       // final QuadCLT []     quadCLTs,
				  fpn,            // final double [][][]  fpn,
				  range,          // final int []         range, // required
				  nsens,          // final int            nsens,
				  fpn_width,      // final int            width,
				  um_sigma,       // final double         um_sigma,
				  max_abs,        // final double         max_abs, //  =        100.0;
				  weight_outlier, // final double         weight_outlier, // = 0.001;
				  dbg_title); // final String         title)
		  if (imp != null) {
			  quadCLT.saveImagePlusInModelDirectory(
					  dbg_title, // String      suffix, // null - use title from the imp
					  imp); // ImagePlus   imp)
			  if (show) {
				  imp.show();
			  }
		  }
		  return imp; // ************** restart from here!
	  }
	
	
	  public static ImagePlus showFPN(
			  double [][][] fpn,
			  int     width,
			  boolean show) {
		  int num_sens = fpn.length;
		  int num_colors = -1; // fpn[0].length;
		  for (int i = 0; i < num_sens; i++) {
			  if (fpn[i] != null) {
				  num_colors = fpn[i].length;
				  break;
			  }
		  }
		  String [] titles = new String[num_sens];
		  String [] top_titles = new String[num_colors];
		  for (int ncol = 0; ncol < num_colors;ncol++) {
			  top_titles[ncol] = "Color-"+ncol;
		  }

		  double [][][] img_data = new double [num_colors][num_sens][];
		  for (int nsens = 0; nsens < num_sens; nsens++) if (fpn[nsens] != null) {
			  titles[nsens]= "Sens-"+nsens;
			  for (int ncol = 0; ncol < num_colors;ncol++) {
				  img_data [ncol][nsens] = fpn[nsens][ncol];
			  }
		  }
		  String fpn_title = "FPN_data";
		  ImagePlus imp = ShowDoubleFloatArrays.showArraysHyperstack(
				  img_data,     // double[][][] pixels, 
				  width,    // int          width, 
				  fpn_title,        // String       title, "time_derivs-rt"+diff_time_rt+"-rxy"+diff_time_rxy,
				  titles,           // String []    titles, // all slices*frames titles or just slice titles or null
				  top_titles,       // String []    frame_titles, // frame titles or null
				  show);            // boolean      show)
		  return imp;
	  }

	  /**
	   * Calculate sensors FPN in a simple way average each channel/color for a scene sequence. Intended to be
	   * used in CUAS mode averaging several (normally 2 full periods of rotation.
	   * TODO: make more accurate and universal by back-propagating the corrected image and subtracting per-sensor
	   * versions before averaging.
	   * @param quadCLTs sequence of scenes with conditioned images (getImageData() != null), without FPN applied
	   * @param range    a first_last scene index pair. May be adjusted to include an integer number of rotations.
	   * @param sensor_mask bitmask which sensors to process (normally -1 - all)
	   * @param debugLevel debug level
	   * @return [sensor][color][pixel] average FPN image. May be saved as image_data to a virtual (center) scene
	   * (a QuadCLT instance). 
	   */
	  public static double [][][] calculateFPN(
			  final QuadCLT []     quadCLTs,
			  final int []         range, // required
			  final int            sensor_mask,
			  final int            debugLevel){
		  QuadCLT first_scene =  quadCLTs[range[0]];
		  final int num_sens =   first_scene.getNumSensors();    	
		  final int num_colors = first_scene.getNumColors();
		  final int width =      first_scene.getTilesX()*first_scene.getTileSize();
		  final int height =     first_scene.getTilesY()*first_scene.getTileSize();
		  final int num_pix =    width*height;
		  final int num_scenes = range[1]-range[0]+1;
		  final double [][][] fpn = new double [num_sens][num_colors][num_pix];
		  final Thread[] threads = ImageDtt.newThreadArray();
		  final AtomicInteger ai = new AtomicInteger(0);
		  for (int ithread = 0; ithread < threads.length; ithread++) {
			  threads[ithread] = new Thread() {
				  public void run() {
					  for (int nChn = ai.getAndIncrement(); nChn < num_sens; nChn = ai.getAndIncrement()) if (((sensor_mask >> nChn) & 1) != 0) {
						  for (int nscene = range[0]; nscene <= range[1]; nscene++) {
							  for (int ncol = 0; ncol < num_colors; ncol++) {
								  double [] img_slice = quadCLTs[nscene].getImageData()[nChn][ncol];
								  for (int npix = 0; npix < num_pix; npix++) {
									  fpn[nChn][ncol][npix] += img_slice[npix];
								  }
							  }
						  }
						  double scale = 1.0/num_scenes;
						  for (int ncol = 0; ncol < num_colors; ncol++) {
							  for (int npix = 0; npix < num_pix; npix++) {
								  fpn[nChn][ncol][npix] *= scale;
							  }
						  }
					  }
				  }
			  };
		  }		      
		  ImageDtt.startAndJoin(threads);
		  return fpn;
	  }
	  
	  public static int getImageHeight(
			  double [][][] image_data,
			  int width) {
		  for (int nChn = 0; nChn < image_data.length; nChn++) {
			  if (image_data[nChn] != null) {
				  return image_data[nChn][0].length/width;
			  }
		  }
		  return 0;
		  
	  }
	  
	  public static double [][]   getRowAvgMulti(
			  final double [][][] image_data,
			  final int           width,
			  final double        max_abs,         // only average within +/- max_abs
			  final double        weight_outlier){ // small weight of outliers to avoid instability
		  final int num_sens = image_data.length;
		  final int height = getImageHeight(
				  image_data, // final double [][][] image_data,
				  width);    // final int width){
		  final double [][] image_row_avg =  new double [num_sens][width];
		  final double [][] weight_row_avg = new double [num_sens][width];
		  final Thread[] threads = ImageDtt.newThreadArray();
		  final AtomicInteger ai = new AtomicInteger(0);
		  for (int ithread = 0; ithread < threads.length; ithread++) {
			  threads[ithread] = new Thread() {
				  public void run() {
					  for (int nChn = ai.getAndIncrement(); nChn < num_sens; nChn = ai.getAndIncrement()) if (image_data[nChn] != null){
						  int indx = 0;
						  for (int y = 0; y < height; y++) {
							  for (int x = 0; x < width; x++) {
								  double d = image_data[nChn][0][indx++];
								  double w = (Math.abs(d) <= max_abs) ? 1.0 : weight_outlier;
								  image_row_avg[nChn][x] +=  w*d;
								  weight_row_avg[nChn][x] += w;
							  }
						  }
						  for (int x = 0; x < width; x++) {
							  image_row_avg[nChn][x] /= weight_row_avg[nChn][x];
						  }
					  }
				  }
			  };
		  }		      
		  ImageDtt.startAndJoin(threads);
		  return image_row_avg;
	  }

	  public static double [][]   getRowAvg(
			  final double [][][] image_data,
			  final int           width,
			  final double        max_abs,         // only average within +/- max_abs
			  final double        weight_outlier){ // small weight of outliers to avoid instability
		  final int num_sens = image_data.length;
		  final int height = getImageHeight(
				  image_data, // final double [][][] image_data,
				  width);    // final int width){
		  final double [][] image_row_avg =  new double [num_sens][width];
		  final double [][] weight_row_avg = new double [num_sens][width];
		  for (int nChn = 0; nChn < num_sens; nChn++) if (image_data[nChn] != null){
			  int indx = 0;
			  for (int y = 0; y < height; y++) {
				  for (int x = 0; x < width; x++) {
					  double d = image_data[nChn][0][indx++];
					  double w = (Math.abs(d) <= max_abs) ? 1.0 : weight_outlier;
					  image_row_avg[nChn][x] +=  w*d;
					  weight_row_avg[nChn][x] += w;
				  }
			  }
			  for (int x = 0; x < width; x++) {
				  image_row_avg[nChn][x] /= weight_row_avg[nChn][x];
			  }
		  }
		  return image_row_avg;
	  }
	  
	  public static double [][] getColAvgMulti(
			  final double [][][] image_data,
			  final int           width,
			  final double        max_abs,         // only average within +/- max_abs
			  final double        weight_outlier){ // small weight of outliers to avoid instability
		  final int num_sens = image_data.length;
		  final int height = getImageHeight(
				  image_data, // final double [][][] image_data,
				  width);    // final int width){
		  final double [][] image_col_avg =  new double [num_sens][height];
		  final double [][] weight_col_avg = new double [num_sens][height];
		  final Thread[] threads = ImageDtt.newThreadArray();
		  final AtomicInteger ai = new AtomicInteger(0);
		  for (int ithread = 0; ithread < threads.length; ithread++) {
			  threads[ithread] = new Thread() {
				  public void run() {
					  for (int nChn = ai.getAndIncrement(); nChn < num_sens; nChn = ai.getAndIncrement()) if (image_data[nChn] != null){
						  int indx = 0;
						  for (int y = 0; y < height; y++) {
							  for (int x = 0; x < width; x++) {
								  double d = image_data[nChn][0][indx++];
								  double w = (Math.abs(d) <= max_abs) ? 1.0 : weight_outlier;
								  image_col_avg[nChn][y] +=  w*d;
								  weight_col_avg[nChn][y] += w;
							  }
						  }
						  for (int y = 0; y < height; y++) {
							  image_col_avg[nChn][y] /= weight_col_avg[nChn][y];
						  }
					  }
				  }
			  };
		  }		      
		  ImageDtt.startAndJoin(threads);
		  return image_col_avg;
	  }
	  
	  public static double [][] getColAvg(
			  final double [][][] image_data,
			  final int           width,
			  final double        max_abs,         // only average within +/- max_abs
			  final double        weight_outlier){ // small weight of outliers to avoid instability
		  final int num_sens = image_data.length;
		  final int height = getImageHeight(
				  image_data, // final double [][][] image_data,
				  width);    // final int width){
		  final double [][] image_col_avg =  new double [num_sens][height];
		  final double [][] weight_col_avg = new double [num_sens][height];
		  for (int nChn = 0; nChn < num_sens; nChn++) if (image_data[nChn] != null){
			  int indx = 0;
			  for (int y = 0; y < height; y++) {
				  for (int x = 0; x < width; x++) {
					  double d = image_data[nChn][0][indx++];
					  double w = (Math.abs(d) <= max_abs) ? 1.0 : weight_outlier;
					  image_col_avg[nChn][y] +=  w*d;
					  weight_col_avg[nChn][y] += w;
				  }
			  }
			  for (int y = 0; y < height; y++) {
				  image_col_avg[nChn][y] /= weight_col_avg[nChn][y];
			  }
		  }
		  return image_col_avg;
	  }
	  
	  
	  
	  
	  public static double [][][] applyRowCol(
			  final double [][][] image_data,
			  final double [][]   image_row_avg,
			  final double [][]   image_col_avg,
			  final boolean       inplace ){
		  final int num_sens = image_data.length;
		  final double [][][] out_data = new double [num_sens][1][];
		  int w=0,h=0;
		  for (int nchn = 0; nchn < num_sens; nchn ++) if (image_data[nchn] != null) {
			  out_data[nchn][0] = inplace ? image_data[nchn][0] : image_data[nchn][0].clone();
			  w = image_row_avg[nchn].length;
			  h = image_col_avg[nchn].length;
		  }
		  final int width =  w;
		  final int height = h;
		  final Thread[] threads = ImageDtt.newThreadArray();
		  final AtomicInteger ai = new AtomicInteger(0);
		  for (int ithread = 0; ithread < threads.length; ithread++) {
			  threads[ithread] = new Thread() {
				  public void run() {
					  for (int nChn = ai.getAndIncrement(); nChn < num_sens; nChn = ai.getAndIncrement()) if (image_data[nChn] != null){
						  int indx = 0;
						  for (int y = 0; y < height; y++) {
							  for (int x = 0; x < width; x++) {
								  out_data[nChn][0][indx++] -=  image_row_avg[nChn][x] + image_col_avg[nChn][y];
							  }
						  }
					  }
				  }
			  };
		  }		      
		  ImageDtt.startAndJoin(threads);
		  return out_data;
	  }

	  public static void applyRowCol( // dangerous - leaves image_row_avg,image_col_avg not set
			  final QuadCLT []     scenes,
			  final double [][][]  rows,
			  final double [][][]  cols ){
		  for (int nscene = 0; nscene < scenes.length; nscene++) {
			  double [][] row = (rows != null) ? rows[nscene] : null;
			  double [][] col = (cols != null) ? cols[nscene] : null;
			  applyRowCol(
					  scenes[nscene], // final QuadCLT       scene,
					  row,            // final double [][]   row,
					  col);           // final double [][]   col )
		  }
		  return;
	  }

	  public static void setApplyRowCol(
			  final QuadCLT []     scenes,
			  final double [][][]  rows,
			  final double [][][]  cols ){
		  for (int nscene = 0; nscene < scenes.length; nscene++) {
			  double [][] row = (rows != null) ? rows[nscene] : null;
			  double [][] col = (cols != null) ? cols[nscene] : null;
			  scenes[nscene].setRowCol(
					  row, // double [][] row,
					  col); // double [][] col)
			  scenes[nscene].applyRowCol();
	//				  scenes[nscene], // final QuadCLT       scene,
    //					  row,            // final double [][]   row,
	//				  col);           // final double [][]   col )
		  }
		  return;
	  }
	  
	  
	  
	  public static void applyRowCol(
			  final QuadCLT       scene,
			  final double [][]   row,
			  final double [][]   col ){
		  if (scene.rowColDiffers(
				  row, // double [][] row,
				  col)) { // double [][] col)) {
			  final double [][] row_applied = scene.getRowApplied();
			  final double [][] col_applied = scene.getColApplied();
			  if (scene.getImageData() == null) {
				  System.out.println("applyRowCol (): image_data==null");
				  return;
			  }
			  final double [][][] image_data = scene.getImageData();
			  final int num_sens = image_data.length;
			  int w=0,h=0;
			  for (int nchn = 0; nchn < num_sens; nchn ++) if (image_data[nchn] != null) {
				  w = row[nchn].length;
				  h = col[nchn].length;
				  break;
			  }
			  final int width =  w;
			  final int height = h;
			  final Thread[] threads = ImageDtt.newThreadArray();
			  final AtomicInteger ai = new AtomicInteger(0);
			  for (int ithread = 0; ithread < threads.length; ithread++) {
				  threads[ithread] = new Thread() {
					  public void run() {
						  for (int nChn = ai.getAndIncrement(); nChn < num_sens; nChn = ai.getAndIncrement()) if (image_data[nChn] != null){
							  if (row_applied!= null) { // unlikely
								  int indx = 0;
								  for (int y = 0; y < height; y++) {
									  for (int x = 0; x < width; x++) {
										  image_data[nChn][0][indx++] += row_applied[nChn][x];
									  }
								  }
							  }
							  if (col_applied!= null) { // unlikely
								  int indx = 0;
								  for (int y = 0; y < height; y++) {
									  for (int x = 0; x < width; x++) {
										  image_data[nChn][0][indx++] += col_applied[nChn][y];
									  }
								  }
							  }
							  if ((row != null) && (col != null)) { // most likely
								  int indx = 0;
								  for (int y = 0; y < height; y++) {
									  for (int x = 0; x < width; x++) {
										  image_data[nChn][0][indx++] -= row[nChn][x] + col[nChn][y];
									  }
								  }
							  } else { // handle case if only one of row, col is nont null (not used so far)
								  if (row != null) {
									  int indx = 0;
									  for (int y = 0; y < height; y++) {
										  for (int x = 0; x < width; x++) {
											  image_data[nChn][0][indx++] -= row[nChn][x];
										  }
									  }
								  }
								  if (col != null) {
									  int indx = 0;
									  for (int y = 0; y < height; y++) {
										  for (int x = 0; x < width; x++) {
											  image_data[nChn][0][indx++] -= col[nChn][y];
										  }
									  }
								  }
								  
							  }
						  }
					  }
				  };
			  }		      
			  ImageDtt.startAndJoin(threads);
//			  quadCLT.setFpnApplied(fpn);
			  scene.setRowColApplied(row, col);
			  scene.setHasNewImageData(true);
		  }
		  return;
	  }
	  
	  
	  
	  
	  
	  
	  public static double [][][][]calculateRowCol(
			  final QuadCLT []     quadCLTs,
			  final double [][][]  fpn, // if null - already applied
			  final int            width,
			  final double         um_sigma,
			  final double         max_abs, //  =        100.0;
			  final double         weight_outlier // = 0.001;
			  ){
		  final int num_sens =     fpn.length;
		  final int num_scenes =  quadCLTs.length; // will use all scenes
		  final int num_pix = fpn[0][0].length;
		  final int height = num_pix / width;
		  final int ncolor = 0;
		  final double [][][] rows = new double[num_scenes][num_sens][width];
		  final double [][][] cols = new double[num_scenes][num_sens][height];
		  final Thread[] threads = ImageDtt.newThreadArray();
		  final AtomicInteger ai = new AtomicInteger(0);
		  for (int ithread = 0; ithread < threads.length; ithread++) {
			  threads[ithread] = new Thread() {
				  public void run() {
					  for (int nScene = ai.getAndIncrement(); nScene < num_scenes; nScene = ai.getAndIncrement()) {
						  double [][][] data_chn = new double [num_sens][][];
						  for (int nsens = 0; nsens < num_sens; nsens++) {
							  double [] img =      quadCLTs[nScene].getImageData()[nsens][ncolor].clone();

							  if ((fpn != null) && (quadCLTs[nScene].getAppliedFPN() != fpn)) {
								  for (int i = 0; i < num_pix; i++) {
									  img[i] -= fpn[nsens][ncolor][i];
								  }
							  }
							  OpticalFlow.applyUMDouble(
									  img, // final double [] data,
									  width,         // final int       width,
									  um_sigma,      // final double um_sigma,
									  1.0);    // final double um_weight)
							  data_chn[nsens] =  new double [][] {img};
						  }
						  rows[nScene] = getRowAvg(
								  data_chn, // final double [][][] image_data,
								  width,             // final int           width,
								  max_abs,           // final double        max_abs,         // only average within +/- max_abs
								  weight_outlier);   // final double        weight_outlier)
						  cols[nScene] = getColAvg(
								  data_chn, // final double [][][] image_data,
								  width,             // final int           width,
								  max_abs,           // final double        max_abs,         // only average within +/- max_abs
								  weight_outlier);   // final double        weight_outlier)
					  }
				  }
			  };
		  }		      
		  ImageDtt.startAndJoin(threads);
		  return new double [][][][] {rows,cols};
	  }
	  
	  
	  public static ImagePlus debugFPN(
			  final QuadCLT []     quadCLTs,
			  final double [][][]  fpn,
			  final int []         range, // required
			  final int            nsens,
			  final int            width,
			  final double         um_sigma,
			  final double         max_abs, //  =        100.0;
			  final double         weight_outlier, // = 0.001;
			  final String         title){
		  
		  int                     ncol = 0;
		  int num_sens = quadCLTs[range[0]].getNumSensors();
		  final double         um_weight = 1.0;
		  int           num_pix = quadCLTs[range[0]].getImageData()[nsens][ncol].length;
		  String [] titles_top = {"src","src-fpn","UM"+um_sigma, "row_col"};
		  double [][][] data = new double [titles_top.length][range[1]-range[0]+1][num_pix];
		  String [] titles = new String [range[1]-range[0]+1];
		  for (int nscene = range[0]; nscene <= range[1]; nscene++) {
			  int iscene = nscene - range[0];
			  titles [iscene] = quadCLTs[nscene].getImageName();
			  data[0][iscene] = quadCLTs[nscene].getImageData()[nsens][ncol].clone();
			  data[1][iscene] = data[0][iscene].clone();  
			  for (int i = 0; i < num_pix; i++) {
				  data[1][iscene][i] -= fpn[nsens][ncol][i];
			  }
			  data[2][iscene] = data[1][iscene].clone();
			  OpticalFlow.applyUMDouble(
					  data[2][iscene], // final double [] data,
					  width,         // final int       width,
					  um_sigma,      // final double um_sigma,
					  um_weight);    // final double um_weight)
			  
			  double [][][] data_chn = new double [num_sens][][];
			  data_chn[nsens] = new double [][] {data[2][iscene]};
			  double [][] image_row_avg = getRowAvgMulti(
					  data_chn, // final double [][][] image_data,
					  width,             // final int           width,
					  max_abs,           // final double        max_abs,         // only average within +/- max_abs
					  weight_outlier);   // final double        weight_outlier)
			  double [][] image_col_avg = getColAvgMulti(
					  data_chn, // final double [][][] image_data,
					  width,             // final int           width,
					  max_abs,           // final double        max_abs,         // only average within +/- max_abs
					  weight_outlier);   // final double        weight_outlier)
			  double [][][] data_out = applyRowCol(
					  data_chn,          // final double [][][] image_data,
					  image_row_avg,     // final double [][]   image_row_avg,
					  image_col_avg,     // final double [][]   image_col_avg,
					  false);            // final boolean       inplace )
			  data[3][iscene] = data_out[nsens][0];
		  }
		  
		  ImagePlus imp = ShowDoubleFloatArrays.showArraysHyperstack(
				  data,        // double[][][] pixels, 
				  width,       // int          width, 
				  title,       // String       title, "time_derivs-rt"+diff_time_rt+"-rxy"+diff_time_rxy,
				  titles,      // String []    titles, // all slices*frames titles or just slice titles or null
				  titles_top,  // String []    frame_titles, // frame titles or null
				  false);      // boolean      show)
		  return imp;
	  }	    
	  
	  
}
